#include "Arduino.h"
#include "ads_combined.h"

#include <bluefruit.h>
#include <string.h>

/* Reads data from Bendlabs One Axis & Two Axis Angular Displacement Sensor using the library
 * ads_combined_driver and sends the data over Bluetooth Low Energy to the Bendlabs Sensor Demo 
 * Android Application.
 * To download the Android companion app go to https://betas.to/2gpFCDTv 
 * 
 * This demo interfaces with the ads sensor in polled mode.
 * SDA   - I2C Data
 * SCL   - I2C Clock
 * VCC   - 3.3 V
 * GND   - Ground
 * nRST  - Not Used
 * nDRDY - Not Used
 */

#define ADS_RESET_PIN       (0xFF)        // Pin number attached to ads reset line. Not used
#define ADS_INTERRUPT_PIN   (0xFF)        // Pin number attached to the ads data ready line. Not used 


BLEService        angms = BLEService(0x1820);
BLECharacteristic angmc = BLECharacteristic(0x2A70);


BLEDis bledis;    // DIS (Device Information Service) helper class instance
BLEBas blebas;    // BAS (Battery Service) helper class instance

// function prototypes
void startAdv(void);
void setupANGM(void);
void connect_callback(uint16_t conn_handle);
void disconnect_callback(uint16_t conn_handle, uint8_t reason);
void ads_data_callback(float * sample);
void deadzone_filter(float * sample);
void signal_filter(float * sample);
void parse_serial_port(void);
 
/* Calback unused, sketch utilizes polled mode */
void ads_data_callback(float * sample)
{

}

#define ADS_DEVICE_TYPE       ADS_ONE_AXIS        // Change to ADS_TWO_AXIS for two axis sensor
#define ADS_XFER_SIZE         (3)                 // Change to 5 for two axis sensor

//ADS Initialization structure
ads_init_t ads_init_s =  { ADS_100_HZ,                //sps
                           &ads_data_callback,        //ads_sample_callback, used for interrupt mode
                           ADS_RESET_PIN,             //reset_pin
                           ADS_INTERRUPT_PIN,         //datardy_pin, used for interrupt mode
                           ADS_DEVICE_TYPE,           //ads_dev_id, ADS_ONE_AXIS, ADS_TWO_AXIS
                           ADS_XFER_SIZE,             //xfer_size, ADS_ONE_AXIS = 3, ADS_TWO_AXIS = 5
                           0                          //addr, 0->default vaue, if ads i2c address is non-default update value
};
  

void setup() {
  Serial.begin(115200);
  Serial.println("ADS BLE Example");
  Serial.println("-----------------------\n");

  Bluefruit.configPrphBandwidth(BANDWIDTH_MAX);

  Bluefruit.begin();
  
  // Set the advertised device name
  Serial.println("Setting Device Name to 'ads_demo'");
  Bluefruit.setName("ads_demo");


  // Set the connect/disconnect callback handlers
  Bluefruit.setConnectCallback(connect_callback);
  Bluefruit.setDisconnectCallback(disconnect_callback);

  Bluefruit.setConnIntervalMS(7.5,30);

  // Configure and Start the Device Information Service
  //Serial.println("Configuring the Device Information Service");
  bledis.setManufacturer("Bend Labs");
  bledis.setModel("One Axis Demo");
  bledis.setHardwareRev("REV: 0.0.0");
  bledis.setSoftwareRev("SD: 132.2.0.0");
  bledis.begin();

  // Start the BLE Battery Service and set it to 100%
  //Serial.println("Configuring the Battery Service");
  blebas.begin();
  blebas.write(100);

  // Setup the Angle Measurement service using
  // BLEService and BLECharacteristic classes
  setupANGM();

  // Setup the advertising packet
  startAdv();
  
  Serial.println("\nAdvertising"); 

  Serial.println("Initializing the Angular Displacement Sensor");                        

  int ret_val = ads_init(&ads_init_s);

  if(ret_val != ADS_OK)
  {
    Serial.print("ADS initialization failed with reason: ");
    Serial.println(ret_val);
  }

  ads_polled(true); // Wake ADS up in polled mode
}

// Command from the characteristic write callback write_callback
static uint8_t write_cb_command = 0xFF;

void loop() {
  // put your main code here, to run repeatedly:
    delay(9);

    float sample[2];

    if(Bluefruit.connected())
    {
      // Read sample from ADS sensor
      if(ads_read_polled(sample) == ADS_OK)
      {
        signal_filter(sample);
        deadzone_filter(sample);

        if(ads_init_s.ads_dev_id == ADS_ONE_AXIS)
        {
          uint8_t ang_encoded[4];
          memcpy(ang_encoded, &sample[0], sizeof(float));
          angmc.notify(ang_encoded, sizeof(ang_encoded));
        }
        else
        {
          uint8_t ang_encoded[8];
          memcpy(ang_encoded, &sample[0], 2*sizeof(float));
          angmc.notify(ang_encoded, sizeof(ang_encoded));
        }
      }

      // Parse received calibration commands from the characteristic write callback
      if(write_cb_command != 0xFF)
      {
        if(ads_init_s.ads_dev_id == ADS_ONE_AXIS)
        {
          if(write_cb_command == 0)
          {
            ads_calibrate(ADS_CALIBRATE_FIRST, 0);
          }
          else if(write_cb_command == 1)
          {
            ads_calibrate(ADS_CALIBRATE_SECOND, 90);
          }
          else if(write_cb_command == 3)
          {
            ads_calibrate(ADS_CALIBRATE_CLEAR, 0);
          }
        }
        else
        {
          if(write_cb_command == 0)
          {
            ads_calibrate(ADS_TWO_CALIBRATE_FIRST, 0);
          }
          else if(write_cb_command == 1)
          {
            ads_calibrate(ADS_TWO_CALIBRATE_FLAT, 90);
          }
          else if(write_cb_command == 2)
          {
            ads_calibrate(ADS_TWO_CALIBRATE_PERP, 90);
          }
          else if(write_cb_command == 3)
          {
            ads_calibrate(ADS_TWO_CALIBRATE_CLEAR, 0);
          }        
        }
        write_cb_command = 0xFF;    // Clear value
      }
    }
}

void signal_filter(float * sample)
{
    static float filter_samples[2][6];

    for(uint8_t i=0; i<2; i++)
    {
      filter_samples[i][5] = filter_samples[i][4];
      filter_samples[i][4] = filter_samples[i][3];
      filter_samples[i][3] = (float)sample[i];
      filter_samples[i][2] = filter_samples[i][1];
      filter_samples[i][1] = filter_samples[i][0];
  
      // 20 Hz cutoff frequency @ 100 Hz Sample Rate
      filter_samples[i][0] = filter_samples[i][1]*(0.36952737735124147f) - 0.19581571265583314f*filter_samples[i][2] + \
        0.20657208382614792f*(filter_samples[i][3] + 2*filter_samples[i][4] + filter_samples[i][5]);   

      sample[i] = filter_samples[i][0];
    }
}

void deadzone_filter(float * sample)
{
  static float prev_sample[2];
  float dead_zone = 0.5f;

  for(uint8_t i=0; i<2; i++)
  {
    if(fabs(sample[i]-prev_sample[i]) > dead_zone)
      prev_sample[i] = sample[i];
    else
      sample[i] = prev_sample[i];
  }
}

void startAdv(void)
{
  // Advertising packet
  Bluefruit.Advertising.addFlags(BLE_GAP_ADV_FLAGS_LE_ONLY_GENERAL_DISC_MODE);
  //Bluefruit.Advertising.addTxPower();

  // Include Angle Measurement Service UUID
  Bluefruit.Advertising.addService(angms);

  // Include Name
  Bluefruit.Advertising.addName();

  Bluefruit.Advertising.addAppearance(113);
  
  /* Start Advertising
   * - Enable auto advertising if disconnected
   * - Interval:  fast mode = 20 ms, slow mode = 152.5 ms
   * - Timeout for fast mode is 30 seconds
   * - Start(timeout) with timeout = 0 will advertise forever (until connected)
   * 
   * For recommended advertising interval
   * https://developer.apple.com/library/content/qa/qa1931/_index.html   
   */
  Bluefruit.Advertising.restartOnDisconnect(true);
  Bluefruit.Advertising.setInterval(500,800);    // in unit of 0.625 ms
  Bluefruit.Advertising.setFastTimeout(30);      // number of seconds in fast mode
  Bluefruit.Advertising.start(0);                // 0 = Don't stop advertising after n seconds  
}

void setupANGM(void)
{
  // Configure the Angle Measurement service
  // Supported Characteristics:
  // Name                         UUID    Requirement Properties
  // ---------------------------- ------  ----------- ----------
  // Angle Measurement            0x2A70  Mandatory   Notify/Write    
  angms.begin();

  // Configure the Angle Measurement characteristic
  // Properties = Notify
  // Min Len    = 4
  // Max Len    = 4
  // Little Endian Float
  angmc.setProperties(CHR_PROPS_NOTIFY|CHR_PROPS_WRITE);
  angmc.setPermission(SECMODE_OPEN, SECMODE_OPEN);
  angmc.setCccdWriteCallback(cccd_callback);  // Optionally capture CCCD updates
  angmc.setWriteCallback(write_callback);
  angmc.begin();
  uint8_t ang_initial[] = {0, 0, 0, 0};
  angmc.notify(ang_initial, 4);                   // Use .notify instead of .write!
}

void connect_callback(uint16_t conn_handle)
{
    Serial.println("Connected");
}

void disconnect_callback(uint16_t conn_handle, uint8_t reason)
{

  Serial.println(reason);
  (void) conn_handle;
  (void) reason;

  Serial.println("Disconnected");
  Serial.println("Advertising!");
}

void write_callback(BLECharacteristic& chr, unsigned char * rx, short unsigned len, short unsigned dah)
{
  if(len == 1)
  {
    write_cb_command = rx[0];
  }
}

void cccd_callback(BLECharacteristic& chr, uint16_t cccd_value)
{
    // Display the raw request packet
    //Serial.print("CCCD Updated: ");
    //Serial.printBuffer(request->data, request->len);
    //Serial.print(cccd_value);
    //Serial.println("");

    // Check the characteristic this CCCD update is associated with in case
    // this handler is used for multiple CCCD records.
    if (chr.uuid == angmc.uuid) {
        if (chr.notifyEnabled()) {
            Serial.println("Angle Measurement 'Notify' enabled");
        } else {
            Serial.println("Angle Measurement 'Notify' disabled");
        }
    }
}
